import scipy
import numpy as np
import pandas as pd


# -------------------------------------------------------------------------
# wl-04-12-2023, Mon: get correlation coefficients, p-values and rt
# differences for compound annotation.
# wl-15-11-2024, Fri: minor changes
# wl-29-07-2025, Sun: call updated 'df_corr_pval'
def comp_corr_rt(df, thres_rt=5.0, thres_corr=0.7, thres_pval=0.05,
                 method="pearson", positive=True):
    """
    Get correlation coefficients, p-values and rt differences for compound
    annotation.

    Parameters
    ----------
    df : DataFrame
        A pandas data frame of peak table. It must have "name", "mz" and
        "rt" columns.
    thres_rt : float
        Threshold for retention time.
    thres_corr : float
        Threshold for correlation coefficients
    thres_pval : float
        Threshold for correlation p-values.
    method : str
        Correlation methods, either "pearson" or "spearman".
    positive : bool
        Use positive correlation or not.

    Returns
    -------
    DataFrame
        A table with differences of retention time, correlation
        coefficients, correlation p-values.
    """

    # get data for correlation analysis
    mat = df.drop(['name', 'mz', 'rt'], axis=1)
    mat = mat.T                      # transpose
    mat.columns = df["name"]         # change columns' names

    # calculate correlation coefficient and p-values
    corr, pval = df_corr_pval(mat, thres_corr=thres_corr,
                              thres_pval=thres_pval, method=method,
                              positive=positive)
    # corr.isnull().sum()    # check nan
    # pval.isnull().sum()    # check nan

    # get rt difference and filter it
    tmp = df[['name', 'rt']]
    diff = abs(df_diff(tmp))
    diff[diff > thres_rt] = np.nan

    # combine three conditions
    ind = corr.notnull() & pval.notnull() & diff.notnull()

    # update
    corr = corr[ind]
    pval = pval[ind]
    diff = diff[ind]

    # convert to long format
    corr = df_short2long(corr).rename(columns={'var': 'r_value'})
    pval = df_short2long(pval).rename(columns={'var': 'p_value'})
    diff = df_short2long(diff).rename(columns={'var': 'rt_diff'})

    # merge
    tab = (
        corr
        .merge(pval, on=['com1', 'com2'], how='inner')
        .merge(diff, on=['com1', 'com2'], how='inner')
        .rename(columns={"com1": "name_a", "com2": "name_b"})
        .round({'r_value': 2})
        .round({'rt_diff': 2})
    )

    return tab


# -------------------------------------------------------------------------
# wl-28-07-2022, Thu: calculate matrix's correlation and p-values
#   This method calls pandas.DataFrame.corr() twice.
# wl-27-09-2022, Tue: get p-values from https://bit.ly/3frFFUV
# wl-05-12-2023, Tue: fix a mis-understanding of sample size (nData)
# wl-06-12-2023, Wed: 'spearman' is extremely slow. Use rank's 'pearson' as
#  the alternative (ultimately doing the same calculation). For details, see
#  https://bit.ly/4a9y1X2
# wl-18-11-2024, Mon: convert float64 to float32
# wl-29-07-2025, Sun: filter corr and pval to reduce memory consumption.
#   (I doubt. Maybe sparse df can be. 02-07-2025, Wed)
def df_corr_pval(df, thres_corr=0.7, thres_pval=0.05, method="pearson",
                 positive=True, min_periods=4):
    """
    Calculate matrix's correlation and p-values.

    This function calls pandas.DataFrame.corr().

    Parameters
    ----------
    df : DataFrame
        A symmetric data frame.
    thres_corr : float
        Threshold for correlation coefficients
    thres_pval : float
        Threshold for correlation p-values.
    method : {'pearson', 'spearman'}
        Method of correlation:
        * pearson : standard correlation coefficient
        * spearman : Spearman rank correlation
    positive : bool
        Use positive correlation or not.
    min_periods : int
        Minimum number of observations required per pair of columns
        to have a valid result.

    Returns
    -------
    corr : DataFrame
        Correlation matrix
    pval : DataFrame
        p-value matrix

    Examples
    --------
    >>> n = 10        # sample size
    >>> m = 20        # feature size
    >>> df = pd.DataFrame(np.random.random((n, m)))
    >>> df.columns = ['col{}'.format(x) for x in range(m)]
    >>> co, pv = df_corr_pval(df)
    """
    def corr_pval_two_side(cc, sample_size):
        # We will divide by 0 if correlation is exactly 1, but that is no
        # problem. We would simply set the test statistic to be infinity if
        # it evaluates to NAN
        with np.errstate(divide='ignore'):
            t = -np.abs(cc) * np.sqrt((sample_size - 2) / (1 - cc**2))
            t[t == np.nan] = np.inf
            t = t.astype('float32')
            # multiply by two to get two-sided p-value
            res = scipy.stats.t.cdf(t, sample_size - 2) * 2
            return res

    df = df_64_32(df)

    # get correlation. Very fast for 'pearson'.
    if False:
        corr = df.corr(method=method, min_periods=min_periods)
    else:
        if method == "pearson":
            corr = df.corr(method="pearson", min_periods=min_periods)
        elif method == "spearman":  # 06-12-2023, Wed: fast than 'spearman'
            corr = df.rank().corr(method="pearson", min_periods=min_periods)
        else:
            raise ValueError("Method must be either 'pearson' or 'spearman'.")

    corr = corr.astype('float32')

    # filter corr
    if positive:
        corr[corr <= thres_corr] = np.nan
    else:
        corr[abs(corr) <= thres_corr] = np.nan

    # get p-value by local function
    # sample size is row of df, not col of df
    pval = corr_pval_two_side(corr, df.shape[0])
    pval = pd.DataFrame(pval)
    # set dimension names as 'corr'
    pval.index = corr.index
    pval.columns = corr.columns

    # filter pval
    pval[pval >= thres_pval] = np.nan
    pval = pval.astype('float32')

    return corr, pval


# -------------------------------------------------------------------------
# wl-20-05-2024, Mon: get correlation group and size
def corr_grp_size(corr):
    """
    Get correlation group and size.

    Parameters
    ----------
    corr : DataFrame
        A long-format correlation table. The front two columns
        names must be 'name_a' and 'name_b'. Other columns could be
        correlation coefficient, p-values and any other values.

    Returns
    -------
    DataFrame
        A data frame with three columns, "name", "cor_grp_size" and
        "cor_grp".
    """
    # get correlated groups/clusters
    cor_nam = corr['name_a'].to_list()
    cor_nam.extend(corr['name_b'].to_list())
    cor_nam = list(set(cor_nam))
    cor_grp = [corr_grp(x, corr) for x in cor_nam]
    # cor_grp_dict = {x: mt.corr_grp(x, corr) for x in cor_nam}
    cor_len = [len(x) for x in cor_grp]
    cor_str = ['::'.join(x) for x in cor_grp]
    # merge into a data frame
    cor_df = pd.DataFrame(list(zip(cor_nam, cor_len, cor_str)),
                          columns=['name', 'cor_grp_size', 'cor_grp'])
    cor_df = cor_df.sort_values('cor_grp_size', ignore_index=True,
                                ascending=False)
    return cor_df


# -----------------------------------------------------------------------
# wl-01-05-2024, Wed: get correlation group
def corr_grp(x, corr):
    """
    Get correlated group.

    'x' will match in the first two columns of 'corr'.

    Parameters
    ----------
    x : str
        A string for match
    corr : DataFrame
        A long-format correlation table. The front two columns
        names must be 'name_a' and 'name_b'. Other columns could be
        correlation coefficient, p-values and any other values.

    Returns
    -------
    list
        A list consisting matched strings.
    """

    ida = corr['name_a'] == x
    a = corr['name_b'][ida].to_list()

    idb = corr['name_b'] == x
    b = corr['name_a'][idb].to_list()

    a.extend(b)

    return a


# -------------------------------------------------------------------------
# wl-04-12-2023, Mon: Convert short to long format and remove NAs
# Note that this function will remove NAs.
def df_short2long(df):
    """
    Convert a symmetric matrix to a pair-wise matrix.

    This function convert so-called short format matrix to long format
    matrix. Also the missing values will be removed.

    Parameters
    ----------
    df : DataFRame
        A symmetric data frame.

    Returns
    -------
    DataFrame
        A data frame.
    """

    # reset columns and index names
    df = df.rename_axis(None).rename_axis(None, axis=1)
    # df.index.name = None

    # reshape data frame
    long_df = df.stack().reset_index()
    long_df.columns = ['com1', 'com2', 'var']

    # Exclude the diagonal and duplicate entries
    if True:
        long_df = long_df[long_df['com1'] < long_df['com2']]
    else:
        # create a mask to identify rows with duplicate features
        # See https://bit.ly/3Ooei9W for details.
        idx = (
            (long_df[['com1', 'com2']].apply(frozenset, axis=1).duplicated()) | \
            (long_df['com1'] == long_df['com2'])
        )
        # update
        long_df = long_df[~idx]

    # reset row names
    long_df.reset_index(drop=True, inplace=True)

    return long_df


# -------------------------------------------------------------------------
# wl-25-07-2022, Mon: calculate pairwise difference
# See https://bit.ly/3outZSs for details.
# wl-18-11-2024, Mon: convert float64 to float32
def df_diff(df):
    """
    Calculate pairwise difference of a data frame.

    Parameters
    ----------
    df : DataFrame
        A data frame.

    Returns
    -------
    DataFrame
        A data from with pair-wise differences.

    Examples
    --------
    >>> tmp = {'Country':['GB','JP','US'],'Values':[20.2,-10.5,5.7]}
    >>> tmp = pd.DataFrame(tmp)
    >>> df_diff(tmp)
    """

    df = df_64_32(df)

    # use apply to get symmetric difference matrix
    dif = df.iloc[:, 1].apply(lambda x: df.iloc[:, 1] - x)
    # dif = abs(dif)

    # change row ad column names
    dif.index = list(df.iloc[:, 0])
    dif.columns = list(df.iloc[:, 0])

    return dif


# -------------------------------------------------------------------------
# wl-18-11-2024, Mon: convert float64 to float32
def df_64_32(df):
    """
    Convert Pandas dataframe's float64 columns to float32.

    Parameters
    ----------
    df : DataFrame
        A data frame with float64 columns.

    Returns
    -------
    DataFrame
        A converted data frame.

    Notes
    -----
    This function is used to large dataset's correlation analysis.
    """
    # 25-06-2025, Wed: add this line otherwise will change in-place
    df = df.copy()
    df[df.select_dtypes(np.float64).columns] = df.select_dtypes(
        np.float64).astype(np.float32)
    return df



# # Quick Start
#
# In this vignette we will demonstrate how to use `LAMP` python package. The
# input data and reference files are located in
# https://github.com/wanchanglin/lamp/tree/master/examples/data.

# ## Setup
#
# To use `LAMP`, the first step is to import some python libraries including
# `LAMP`.

import sqlite3
import pandas as pd
from lamp import anno, stats, utils

# ## Data Loading
#
# `LAMP` supports text files separated by comma (`,`) or tab (`\t`).  The
# Microsoft's XLSX is also supported, using argument `sheet_name` to
# indicate which sheet is used for input data. The default is 0 for the
# first sheet.
#
# Here we use a small example data set with `tsv` format. Load it into
# python and check its format:
#

d_data = "./data/df_pos_2.tsv"
data = pd.read_table(d_data, header=0, sep="\t")
data

# This data set includes peak list and intensity data matrix. `LAMP`
# requires peak list's name, m/z value and retention time. User needs to
# indicate the locations of feature name, m/z value, retention time and
# starting points of data matrix from data. Here they are 1, 3, 6 and 11,
# respectively.
#
# Load input data with `xlsx` format for `LAMP`:

cols = [1, 3, 6, 11]
# d_data = "./data/df_pos_2.tsv"
# df = anno.read_peak(d_data, cols, sep='\t')
d_data = "./data/df_pos_2.xlsx"                      # use xlsx file
df = anno.read_peak(d_data, cols, sheet_name=0)
df

# The argument `sep` will be ignored if the input data is an `xlsx` file.
# Data frame `df` now includes only `name`, `mz`, `rt` and intensity data
# matrix.
#
# ## Metabolite Annotation
#
# To perform metabolite annotation, users should provide their own
# reference file. Otherwise, `LAMP` will use its default reference file for
# annotation. Here we load the default reference file for compound
# annotation. Since the input data is positive mode here, we only
# use positive part of reference file. If `ion_mode` is empty, all reference
# items will be used for matching.

ion_mode = "pos"
ref_path = ""  # if empty, use default reference file for matching
# load reference library
cal_mass = False
ref = anno.read_ref(ref_path, ion_mode=ion_mode, calc=cal_mass)
ref

# The reference file must have one column: `molecular_formula` (or
# `formula`) if there is no column called `ion m/z` (or, `m/z`,
# `exact_mass`). The `exact_mass` is optional. if absent, `LAMP` will use
# `molecular_formula` to calculate 'exact_mass' based on the NIST Atomic
# Weights and Isotopic Compositions for All Elements. If your reference file
# has `exact_mass` and you still want to calculate it using NIST database,
# set `calc` as True.  The `exact_mass` is used to match against a range of
# `mz`, controlled by `ppm`, in data frame `df`.
#
# As the same as input data, the reference file can be `xlsx` file. Another
# reference file is HMDB database for urine:

ref_path = "./data/hmdb_urine_v4_0_20200910_v1.tsv"
ref = anno.read_ref(ref_path, calc=True)
ref

# Next we use HMDB reference file for compounds match. Here function argument
# `ppm` is used to control the m/z matching tolerance(range).

ppm = 5.0
match = anno.comp_match_mass(df, ppm, ref)
match

# `match` gives the compound matching results. `LAMP` also provides a mass
# adjust option by adduct library. You can provide your own adducts library
# otherwise `LAMP` uses its default adducts library.
#
# The adducts library's format looks like:

add_path = './data/adducts_short.tsv'
lib_df = pd.read_csv(add_path, sep="\t")
lib_df

# The adducts library must have columns of `label`, `exact_mass`, `charge`
# and `ion_mode`.
#
# We use this adducts file to adjust mass:

# if empty, use default adducts library
add_path = "./data/adducts_short.tsv"
lib_add = anno.read_lib(add_path, ion_mode)
lib_add

# Now use function `comp_match_mass_add` to match compounds:

match_1 = anno.comp_match_mass_add(df, ppm, ref, lib_add)
match_1

# Note that this adducts library is also used to adjust mass calculation in
# loading reference file if there is a column called `ion_type`.
#
# ## Correlation Analysis
#
# Next step is correlation analysis, based on intensity data matrix along all
# peaks. All results are filtered by the correlation coefficient, p-values
# and retention time difference. That is: keep correlation results in an
# retention time differences/window (such as 1 second) with correlation
# coefficient larger than a threshold (such as 0.5) and their correlation
# p-values less than a threshold (such as 0.05).
#
# `LAMP` supports two correlation methods, `pearson` and `spearman`. Also
# parameter `positive` allows user to select only positive correlation
# results, otherwise positive and negative correlations will be used.
#
# Two functions, `_tic` and `_toc`, record the correlation computation time in
# seconds.

thres_rt = 1.0
thres_corr = 0.5
thres_pval = 0.05
method = "spearman"  # "pearson"
positive = True

utils._tic()
corr = stats.comp_corr_rt(df, thres_rt, thres_corr, thres_pval, method,
                          positive)
utils._toc()
corr

# `corr` gives results of correlation coefficient(`r_value`), correlation
# p-values(`p_value`) and retention time difference(`rt_diff`).
#
# Based on the correlation analysis, we can extract the groups and their
# sizes by:

# get correlation group and size
corr_df = stats.corr_grp_size(corr)
corr_df

# ## Summarize Results
#
# The final step gets the summary table in different format and save for the
# further analysis.

# get summary of metabolite annotation
sr, mr = anno.comp_summ(df, match)

# This function combines peak table with compound matching results and
# returns two results in different formats. `sr` is single row results for
# each peak id in peak table `df`:

sr

# `mr` is multiple rows format if the match more than once from the reference
# file:

mr

#
# Now we merges single format results with correlation results:
#

# merge summery table with correlation analysis
res = anno.comp_summ_corr(sr, corr_df)
res

# The result data frame `res` is re-arranged as four parts from top to bottom:
#
#  - 1st part: identified metabolites, satisfied with correlation analysis
#  - 2nd part: identified metabolites, not satisfied with correlation
#  - 3rd part: no identified metabolites, satisfied with correlation
#  - 4th part: no identified metabolites, not satisfied with correlation
#
# The users should focus on the first part and perform their further analysis.

# You can save all results in different forms, such as text format TSV or CSV.
# You can also save all results into a `sqlite3` database and use
# [DB Browser for SQLite](https://sqlitebrowser.org/) to view:

f_save = False          # here we do NOT save results
db_out = "test.db"
sr_out = "test_s.tsv"

if f_save:
    # save all results into a sqlite3 database
    conn = sqlite3.connect(db_out)
    df[["name", "mz", "rt"]].to_sql("peaklist",
                                    conn,
                                    if_exists="replace",
                                    index=False)
    corr_df.to_sql("corr_grp", conn, if_exists="replace", index=False)
    corr.to_sql("corr_pval_rt", conn, if_exists="replace", index=False)
    match.to_sql("match", conn, if_exists="replace", index=False)
    mr.to_sql("anno_mr", conn, if_exists="replace", index=False)
    res.to_sql("anno_sr", conn, if_exists="replace", index=False)

    conn.commit()
    conn.close()

    # save final results
    res.to_csv(sr_out, sep="\t", index=False)

# ## End User Usages
#
# For end users, `LAMP` provides two computation options: command line
# interface(CLI) and graphical user interface (GUI).
#
# To use GUI,  you need to open a terminal and type in:
#
# ```bash
# $ lamp gui
# ```
#
# To use CLI, open a terminal and type in command with required arguments,
# something like:
#
# ```bash
# $ lamp cli \
#   --input-data "./data/df_pos_3.tsv" \
#   --sep "tab" \
#   --col-idx "1, 2, 3, 4" \
#   --add-path "" \
#   --ref-path "" \
#   --ion-mode "pos" \
#   --cal-mass \
#   --thres-rt "1.0" \
#   --thres-corr "0.5" \
#   --thres-pval "0.05" \
#   --method "pearson" \
#   --positive \
#   --ppm "5.0" \
#   --save-db \
#   --save-mr \
#   --db-out "./res/test.db" \
#   --sr-out "./res/test_s.tsv" \
#   --mr-out "./res/test_m.tsv"
# ```
#
# For the best practice, you can create a bash script `.sh` (Linux
# and MacOS) or Windows script `.bat` to contain these CLI
# arguments. Change parameters in these files each time when processing new
# data set.
#
# For example, there are `lamp_cli.sh` and `lamp_cli.bat` in
# https://github.com/wanchanglin/lamp/tree/master/examples. You can run them
# and check the results in directory `examples/res`:
#
# - For Linux and MacOS terminal:
#
#   ```bash
#   $ chmod +x lamp_cli.sh
#   $ ./lamp_cli.sh
#   ```
#
# - For Windows terminal:
#
#   ```bash
#   $ lamp_cli.bat
#   ```
#
# Note that if users use `xlsx` files for input data and reference file
# when using GUI or CLI, all data must be in the first sheet. If you use
# `LAMP` functions in your python scripts, there are no such requirementss.



import os
import re
import sqlite3
import pandas as pd
import numpy as np
import janitor      # remove_empty()
from pyteomics import mass
from collections import OrderedDict
import lamp
from lamp.utils import (df2dict, flatten_cols, flatten_list)


# ------------------------------------------------------------------------
# wl-08-10-2024, Tue: merge single row summary with correlation analysis
def comp_summ_corr(sr, corr_df):
    """
    Merge annotation table with correlation analysis

    Parameters
    ----------
    sr : DataFrame
        A pandas data frame of annotation summary with single row and
        multiple column format.
    corr_df : DataFrame
        A pandas data frame of correlation group and its count.

    Returns
    -------
    DataFrame
        A data frame of combined annotation summary table.
    """

    # merge summery table with correlation analysis
    res = (
        sr.merge(corr_df, left_on="name", right_on='name', how="left")
        .sort_values(["cor_grp_size"], ignore_index=True, ascending=False)
    )

    # sort out based on correlation group and compound matches
    idx = res.ppm_error.notnull() & res.cor_grp_size.notnull()
    # split into two groups
    res_1 = res[idx]
    res_2 = res[~idx].sort_values(["ppm_error"])

    res = pd.concat([res_1, res_2], ignore_index=True)

    return res


# ------------------------------------------------------------------------
# wl-14-05-2024, Tue: get summary table of compound annotation
def comp_summ(pk, comp, unique=False):
    """
    Get annotation summary table.

    Parameters
    ----------
    pk : DataFrame
        A pandas data frame of peak table. It must have "name", "mz" and
        "rt" columns.
    comp : DataFrame
        A pandas data frame of metabolite annotation.
    unique : bool
        A flag for using unique match or not

    Returns
    -------
    sr_mc : DataFrame
        A pandas dataframe with single row and multiple column table.
    mr_mc : DataFrame
        A pandas dataframe with multiple row and multiple column table.
    """

    mr_mc = (
        pk[["name", "mz", "rt"]]
        .merge(comp, left_on="name", right_on="id", how="outer",
               suffixes=('', '_remove'))
        .loc[:, lambda x: [i for i in x.columns if 'remove' not in i]]
        .drop(["id"], axis=1)
        # .sort_values(["name", "compound_id"], ignore_index=True)
    )

    ret = comp_merge(comp, unique=unique)
    sr_mc = (
        pk[["name", "mz", "rt"]]
        .merge(ret, left_on="name", right_on="name", how="outer",
               suffixes=('', '_remove'))
        .loc[:, lambda x: [i for i in x.columns if 'remove' not in i]]
    )

    return sr_mc, mr_mc


# ------------------------------------------------------------------------
# wl-20-05-2024, Mon: merge compound annotation
# wl-14-08-2024, Wed: give explanation of this function.
# wl-22-08-2024, Thu: add a flag for unique result or not
# wl-17-09-2025, Wed: add 'total_matching' for non-unique result
def comp_merge(comp, unique=False):
    """
    Merge compound annotation.

    This function merges all metabolite matches of a peak in single
    row for summary table.

    Parameters
    ----------
    comp : DataFrame
        A pandas data frame of metabolite match.
    unique : bool
        A flag for using unique match or not

    Returns
    -------
    DataFrame
        A pandas dataframe with merged columns.
    """
    # --------------------------------------------------------------------
    # Join string from a list
    # join_str(['ab', 'ab', 'efg', 'cd', 'efg', 'cd', 'efg', 'cd', 'efg'])
    def join_str(s):
        # s = [x for x in s if x]
        return '::'.join(s)

    # --------------------------------------------------------------------
    # Join unique string from a list
    # uni_str(['ab', 'ab', 'efg', 'cd', 'efg', 'cd', 'efg', 'cd', 'efg'])
    def uni_str(s):
        # s = [x for x in s if x]
        return '::'.join(list(set(s)))

    # -------------------------------------------------------------------
    # Count unique string from a list
    # uni_count(['ab', 'ab', 'efg', 'cd', 'efg', 'cd', 'efg', 'cd', 'efg'])
    def uni_count(s):
        # s = [x for x in s if x]
        return len(list(set(s)))

    # Get unique string and count or not
    if unique:
        tmp = (
            comp
            .select_dtypes(include='object')      # only select string
            .groupby("id")
            .agg([uni_str, uni_count])
            .pipe(flatten_cols)
            .rename(columns=lambda x: x.replace('_uni_str', ''))
            .rename(columns=lambda x: x.replace('_uni', ''))
            .rename_axis("name").reset_index()    # add rowname as a column
        )
    else:
        tmp_1 = (
            comp
            .select_dtypes(include='object')      # only select string
            # .drop("mz", axis=1)
            .groupby("id")
            # .groupby("id", as_index=False)
            .agg(join_str)
            .rename_axis("name").reset_index()    # add row-name as a column
        )
        # wl-17-09-2025, Wed: add matching count using apply on groupby
        tmp_2 = (
            comp
            .select_dtypes(include='object')      # only select string
            .groupby("id", group_keys=False)
            .apply(lambda x: x.shape[0], include_groups=False)
            .rename_axis("name").reset_index()    # add row-name as a column
            .rename(columns={0: "total_match"})
        )
        # merge two parts
        tmp = tmp_1.merge(tmp_2, on="name", how="left")

    # Get non-string columns
    # wl-15-08-2024, Thu: fix a bug
    num = list(comp.select_dtypes(exclude=['object']).columns)
    num.insert(0, "id")
    tmp1 = (
        comp
        .drop_duplicates('id', ignore_index=True)
        .loc[:, num]
        .rename(columns={'id': 'name'})
    )

    # merge two parts
    res = tmp1.merge(tmp, on="name", how="left")

    # Only select count columns for screen and return it?
    # tmp.filter(regex='_count$|^name', axis=1)

    return res


# ------------------------------------------------------------------------
# wl-14-05-2024, Tue: get summary of compound annotation
def _comp_summ_1(pk, comp):
    """
    Get annotation summary table.

    Parameters
    ----------
    pk : DataFrame
        A pandas data frame of peak table. It must have "name", "mz" and
        "rt" columns.
    comp : DataFrame
        A pandas data frame of metabolite annotation.

    Returns
    -------
    sr_mc : DataFrame
        A pandas dataframe with single row and multiple column table.
    mr_mc : DataFrame
        A pandas dataframe with multiple row and multiple column table.

    See Also
    --------
    comp_summ : Return two summary tables.

    Notes
    -----
    This is another version for metabolite annotation. And it does not call
    function `comp_merge`.
    """

    pk = pk[["name", "mz", "rt"]]

    mr_mc = (
        pk.merge(comp, left_on="name", right_on="id", how="outer",
                 suffixes=('', '_remove'))
        .loc[:, lambda x: [i for i in x.columns if 'remove' not in i]]
        .drop(["id"], axis=1)
        # .sort_values(["name", "compound_id"], ignore_index=True)
    )

    com = mr_mc.columns.to_list()
    com = [x for x in com if x not in
           ['name', 'mz', 'rt', 'exact_mass', 'ppm_error']]

    sr_mc = mr_mc.copy()

    sr_mc = (
        sr_mc
        # .fillna("")
        .assign(**{c: lambda x, y=c: x[y].astype(str) for c in com})
        .assign(
            **{c: lambda x, y=c: x.groupby("name")[y].transform("::".join)
               for c in com}
        )
        .drop_duplicates("name")
        .sort_values("name", ignore_index=True)
    )

    return sr_mc, mr_mc


# -------------------------------------------------------------------------
# wl-29-04-2024, Mon: Compound match without adducts library
# wl-06-08-2024, Tue: The only requirement from 'ref' is 'exact_mass'
def comp_match_mass(peak, ppm, ref):
    """
    Compound match.

    This function performs compound match against exact mass in reference
    library.

    Parameters
    ----------
    peak : DataFrame
        A pandas data frame of peak table. It must have "name", "mz" and
        "rt" columns.
    ppm : float
        A value for ppm.
    ref : DataFrame
        A pandas data frame of a library which must have `exact_mass` column.

    Returns
    -------
    DataFrame
        A pandas dataframe of compound match table.

    See Also
    --------
    com_match_mass_add : Return compound match table. This function uses a
        adducts library to adjust compound match.
    """

    # ---------------------------------------------------------------------
    # wl-29-04-2024, Mon: select compounds based on exact mass
    def comp_sel_mass(tab_name, col_names, cur, peak_id, mz, ppm):
        min, max = _cal_mass_tol(mz, ppm)
        rec = []
        cur.execute(
            """
            SELECT * from {} where exact_mass >= {} and exact_mass <= {}
            """.format(tab_name, min, max)
        )
        rec = [OrderedDict(zip(col_names, list(record)))
               for record in cur.fetchall()]
        for record in rec:
            record["id"] = peak_id
            record["mz"] = mz
            record["ppm_error"] = (mz - record["exact_mass"]) / (
                record["exact_mass"] * 0.000001
            )
        return rec

    # ---------------------------------------------------------------------
    # convert df to sql for speedy match
    con = sqlite3.connect(":memory:")
    cur = con.cursor()
    ref.to_sql(name='ref', con=con, if_exists="replace", index=False)
    con.commit()

    cur.execute("SELECT name FROM sqlite_master WHERE type='table'")
    tab_name = list(cur.fetchall())[0][0]
    # set index for fast query
    cur.execute(
        "CREATE INDEX idx_exact_mass ON {} (exact_mass)".format(tab_name))

    # get table column names
    if True:
        col_names = ref.columns.to_list()
    else:
        cur.execute("SELECT * FROM {}".format(tab_name))
        col_names = list(map(lambda x: x[0], cur.description))

    # convert peak list to dictionary for query
    pk = df2dict(peak[["name", "mz"]])

    # annotation/match compounds
    res = [comp_sel_mass(tab_name, col_names, cur, x, pk[x], ppm)
           for x in pk]
    res = flatten_list(res)
    res = pd.DataFrame(res)

    res = (
        res
        .round({"ppm_error": 2, "exact_mass": 2})
        # .sort_values(["id", "compound_id"], ignore_index=True)
        .drop_duplicates(ignore_index=True)
    )

    # move id and mz at the front
    cols_to_move = ['id', 'mz']
    res = res[cols_to_move +
              [x for x in res.columns if x not in cols_to_move]]

    # wl-17-09-2025, Wed: convert all ref as string
    cols = ref.columns.to_list()
    res = res.assign(**{c: lambda x, y=c: x[y].astype(str) for c in cols})

    con.close()

    return res


# -------------------------------------------------------------------------
# wl-24-04-2024, Wed: compound match
# Convert data frame 'ref' to sqlite for speedy query
# wl-06-08-2024, Tue: The only requirement from 'ref' is 'exact_mass'
# wl-07-08-2024, Wed: 'adduct' in peak has nothing to do with lib_adducts.
def comp_match_mass_add(peak, ppm, ref, lib_adducts):
    """
    Compound match.

    This function performs compound match against exact mass in reference
    library and adducts library.

    Parameters
    ----------
    peak : DataFrame
        A pandas data frame of peak table. It must have "name", "mz" and
        "rt" columns.
    ppm : float
        A value for ppm.
    ref : DataFrame
        A pandas data frame of reference library which must have
        `exact_mass` column.
    lib_adducts : DataFrame
        A pandas data frame of adducts library which must have `exact_mass`
        column.

    Returns
    -------
    DataFrame
        A pandas dataframe of compound match table.

    See Also
    --------
    com_match_mass : Return compound match table. This function does not use
        adducts library to adjust compound match.
    """

    # --------------------------
    # convert df to sql for speedy match
    con = sqlite3.connect(":memory:")
    cur = con.cursor()
    ref.to_sql(name='ref', con=con, if_exists="replace", index=False)
    con.commit()

    cur.execute("SELECT name FROM sqlite_master WHERE type='table'")
    tab_name = list(cur.fetchall())[0][0]
    # set index for fast query
    cur.execute(
        "CREATE INDEX idx_exact_mass ON {} (exact_mass)".format(tab_name))

    # get table column names
    if True:
        col_names = ref.columns.to_list()
    else:
        cur.execute("SELECT * FROM {}".format(tab_name))
        col_names = list(map(lambda x: x[0], cur.description))

    # --------------------------
    # convert peak list to dictionary for query
    pk = df2dict(peak[["name", "mz"]])
    # pk = peak.set_index('name').T.to_dict('records')[0]

    # convert lib_adducts to dictionary for query
    lib = (
        lib_adducts
        .rename(columns={"exact_mass": "mass"})
        # .set_index('label').T.to_dict('list')
        .set_index('label').T.to_dict('dict')
    )

    # annotation/match compounds
    res = [_comp_sel(lib, tab_name, col_names, cur, x, pk[x], ppm)
           for x in pk]
    res = flatten_list(res)
    res = pd.DataFrame(res)

    # clean up
    res = (
        res
        .round({"ppm_error": 2, "exact_mass": 2})
        # .sort_values(["id", "compound_id"], ignore_index=True)
        # .drop_duplicates(['id', 'compound_id'], ignore_index=True)
        .drop_duplicates(ignore_index=True)
    )

    # move id and mz at the front
    cols_to_move = ['id', 'mz']
    res = res[cols_to_move +
              [x for x in res.columns if x not in cols_to_move]]

    con.close()

    return res


# -------------------------------------------------------------------------
# wl-11-01-2024, Thu: select compounds based on exact mass
# wl-24-04-2024, Wed: use all column names
def _comp_sel(lib, tab_name, col_names, cur, peak_id, mz, ppm):
    """Internal function for `comp_match_mass_add`."""

    min_mz, max_mz = _cal_mass_tol(mz, ppm)
    adducts = lib.keys()

    rec = []
    for adduct in adducts:    # adduct = '[M+Mg]+'
        if min_mz - lib[adduct]["mass"] < 0.5:
            continue

        min = (min_mz - lib[adduct]["mass"]) * lib[adduct]["charge"]
        max = (max_mz - lib[adduct]["mass"]) * lib[adduct]["charge"]

        cur.execute(
            """
            SELECT * from {} where exact_mass >= {} and exact_mass <= {}
            """.format(tab_name, min, max)
        )

        sub = [OrderedDict(zip(col_names, list(record)))
               for record in cur.fetchall()]

        for record in sub:
            record["exact_mass"] = (
                record["exact_mass"] / lib[adduct]["charge"]
                + lib[adduct]["mass"]
            )
            record["adduct"] = adduct

        rec.extend(sub)

    for record in rec:
        record["id"] = peak_id
        record["mz"] = mz
        record["ppm_error"] = (mz - record["exact_mass"]) / (
            record["exact_mass"] * 0.000001
        )

    return rec


# -------------------------------------------------------------------------
# wl-22-04-2024, Mon: calculate mass tolerance/range based on ppm
# wl-11-10-2024, Fri: Review
def _cal_mass_tol(mass, ppm):
    """
    Calculate m/z value tolerance/range.

    Parameters
    ----------
    mass : float
        A mass value.
    ppm : float
        ppm value.

    Returns
    -------
    min_tol : float
        Minimal tolerance.
    max_tol : float
        Maximal tolerance.
    """

    delta = 0.000001
    min_tol = (1 - delta * ppm) * mass
    max_tol = (1 + delta * ppm) * mass

    return min_tol, max_tol


# -----------------------------------------------------------------------
# wl-22-04-2024, Mon: calculate molecular formula' exact mass
# wl-23-04-2024, Tue: fix a bug. use argument names for 'calculate_mass'
# wl-06-08-2024, Tue: exception handling. how to catch user-defined error:
#  PyteomicsError?
def _cal_mass(formula, mass_data):
    """
    Calculate molecular formula's exact mass

    Parameters
    ----------
    formula : str
        A string with a chemical formula.
    mass_data : OrderedDict
        A dict with the masses of the chemical elements.

    Returns
    -------
    float
        A mass value
    """

    formula = str(formula)
    try:
        res = mass.calculate_mass(formula=formula, mass_data=mass_data)
        # res = round(res, 6)
    # except (KeyError, PyteomicsError, NameError):
    except Exception:
        res = np.nan

    return res


# -----------------------------------------------------------------------
# wl-08-08-2024, Thu: adjust molecular formula' exact mass
# wl-06-11-2024, Wed: Re-write
def _adj_mass(mass, adduct, add_dict):
    """
    Adjust mass based on adduct library

    Parameters
    ----------
    mass : float
        Mass value to be adjusted.
    adduct : str
        A string of adduct. It is used to match against 'add_dict'.
    add_dict : dict
        A dictionary of adducts library.

    Returns
    -------
    float
        Adjusted mass
    """

    if adduct in add_dict:
        adj = mass + add_dict[adduct]
    else:
        adj = mass

    return adj


# -----------------------------------------------------------------------
# wl-22-04-2024, Mon: calculate exact mass for a reference data frame
# wl-07-08-2024, Wed: problem if adduct does not match in adduct library
# wl-08-08-2024, Thu: fix a bug in mass adjustment
def cal_mass(df, lib_adducts=None):
    """
    Calculate exact mass for an reference data frame.

    This function calculates exact mass using NIST database. If the
    reference file has 'adduct' column and user provides a adducts library,
    the exact mass will be adjust according to the adducts library.

    Parameters
    ----------
    df : DataFrame
        An reference data frame. It must have a column called
        'molecular_formula'.
    lib_adducts : DataFrame
        Adducts library in data frame format for mass adjustment. Only for
        'df' has 'ion_type' column.

    Returns
    -------
    DataFrame
        Reference dataframe with exact mass.
    """

    # --------------------------------------------------------------------
    def parse_nist_database(fn, skip_lines=10):
        """
        Parse NIST database.

        This function is called by `nist_database_to_pyteomics`.

        Parameters
        ----------
        fn : str
            Text file (NISTs Linearized ASCII Output)
        skip_lines : integer
            The number of lines of the data file to skip before beginning to
            read data.

        Yields
        ------
        Ordered dictionary
            Containing the parsed records

        Notes
        -----
        This function is from python package `beamspy`.
        """

        with open(fn, "r") as inp:
            for i in range(skip_lines):
                inp.readline()
            for e in inp.read().split("\n\n"):
                record = OrderedDict()
                for line in e.strip().split("\n"):
                    kv = line.split(" =")
                    if kv[0] == "Relative Atomic Mass":
                        record[kv[0]] = re.findall(r'\d+(?:\.\d+)?', kv[1])
                        record[kv[0]][0] = float(record[kv[0]][0])
                        record[kv[0]][1] = int(record[kv[0]][1])

                    elif kv[0] == "Isotopic Composition":
                        matches = re.findall(r'\d+(?:\.\d+)?', kv[1])
                        if len(matches) > 0:
                            record[kv[0]] = matches
                            if len(matches) > 1:
                                record[kv[0]][0] = float(record[kv[0]][0])
                                record[kv[0]][1] = int(record[kv[0]][1])
                            else:
                                record[kv[0]] = [float(record[kv[0]][0]), None]
                        else:
                            record[kv[0]] = [0.0, None]
                    elif kv[0] == "Atomic Number" or kv[0] == "Mass Number":
                        record[kv[0]] = int(kv[1])
                    elif kv[0] == "Standard Atomic Weight":
                        matches = re.findall(r'\d+(?:\.\d+)?', kv[1])
                        record[kv[0]] = matches
                    else:
                        record[kv[0]] = kv[1].strip()
                yield record

    # --------------------------------------------------------------------
    def nist_database_to_pyteomics(fn, skip_lines=10):
        """
        Convert NIST database to dict for calculation of mass

        Parameters
        ----------
        fn : str
            Text file (NISTs Linearized ASCII Output)
        skip_lines : integer
            The number of lines of the data file to skip before beginning to
            read data.

        Returns
        -------
        dict
            Ordered dictionary containing NIST records compatible with
            `Pyteomics`.

        Notes
        -----
        This function is from python package `beamspy`.
        """

        def add_record(r, nm):
            if r["Atomic Symbol"] not in nm:
                # update after all records have been added
                nm[r["Atomic Symbol"]] = OrderedDict([(0, (0.0, 0.0))])
                nm[r["Atomic Symbol"]][r["Mass Number"]] = (
                    r["Relative Atomic Mass"][0], r["Isotopic Composition"][0])
            else:
                nm[r["Atomic Symbol"]][r["Mass Number"]] = (
                    r["Relative Atomic Mass"][0], r["Isotopic Composition"][0])
            return nm

        def order_composition_by_hill(composition):
            symbols = set(composition)
            if 'C' in symbols:
                symbols.remove('C')
                yield 'C'
                if 'H' in symbols:
                    symbols.remove('H')
                    yield 'H'
            for symbol in sorted(symbols):
                yield symbol

        lib = OrderedDict()
        for record in parse_nist_database(fn, skip_lines=skip_lines):
            if record["Atomic Symbol"] in ["D", "T"]:
                lib = add_record(record, lib)
                record["Atomic Symbol"] = "H"
                lib = add_record(record, lib)
            else:
                lib = add_record(record, lib)

        for element in list(lib.keys()):
            lib_sorted = sorted(lib[element].items(),
                                key=lambda e: e[1][1], reverse=True)
            if lib_sorted[0][1][0] > 0.0:
                lib[element][0] = (lib_sorted[0][1][0], 1.0)
            elif len(lib_sorted) == 2:
                lib[element][0] = (lib_sorted[1][1][0], 1.0)
            else:
                del lib[element]

        es = list(order_composition_by_hill(lib.keys()))

        return OrderedDict((k, lib[k]) for k in es)

    # --------------------------------------------------------------------
    nist_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'lib', 'nist_database.txt')
    nist_mass = nist_database_to_pyteomics(nist_path)

    # get exact mass
    df = (
        df
        # wl-08-08-2024, Thu: use Series's apply here for single columns
        .assign(exact_mass=lambda x:
                x.molecular_formula.apply(_cal_mass, mass_data=nist_mass))
    )

    # adjust exact mass or not
    if ("adduct" in df.columns) and (lib_adducts is not None):
        # convert 2-column df into dictionary for speedy query
        add = df2dict(lib_adducts.iloc[:, 0:2])
        df = (
            df
            # wl-08-08-2024, Thu: use df's apply here with multiple columns
            #  and cell-wise custom function.
            .assign(exact_mass=lambda x: x.apply(
                lambda x: _adj_mass(x.exact_mass, x.adduct, add), axis=1))
        )

    return df


# ------------------------------------------------------------------------
# wl-22-04-2024, Mon: load reference file for compound annotation
# wl-08-08-2024, Thu: use 'cal_mass' function
# wl-10-09-2024, Tue: remove empty rows and columns
# wl-11-09-2024, Wed: rename 'name' as 'molecular_name' in case 'name'
#   conflicts with data set's peak 'name'.
# wl-09-10-2024, Wed: set 'filename' default value
# wl-23-10-2024, Wed: support Excel
# wl-30-10-2024, Wed: add ion_mode
def read_ref(fn="", ion_mode="pos", sheet_name=0, sep="\t", calc=False,
             lib_adducts=None):
    """
    Load reference for compound annotation

    Parameters
    ----------
    fn : str
        Full path for an reference file. If empty, use default reference
        file. This file must have `formula` or `molecular_formula` column.
        The current supported file formats are text formats (csv, tsv, txt
        and dat) and Excel formats (xls, xlsx).
    ion_mode : str
        A string for ion mode, "pos" or "neg".
    sheet_name: str or int
        This argument is used for Excel file format. Strings are used for
        sheet names. Integers are used in zero-indexed sheet positions
        (chart sheets do not count as a sheet position).

        Available cases:

        * Defaults to ``0``: 1st sheet as a `DataFrame`
        * ``1``: 2nd sheet as a `DataFrame`
        * ``"Sheet1"``: Load sheet with name "Sheet1"

    sep : str
        Character to treat as the delimiter. It is used for text file
        format.
    calc : bool
        Calculate exact mass or not.
    lib_adducts : DataFrame
        Adducts library in data frame format for mass adjustment. Only for
        'fn' has 'ion_type' column.

    Returns
    -------
    DataFrame
        Reference dataframe with exact mass.

    Notes
    -----
    This function will remove reference empty rows and columns.
    """

    # load default reference library
    if not fn:
        path = 'lib/GSMM_LAMP_ReferenceFile_v1_281024.xlsx'
        fn = os.path.join(
            os.path.dirname(os.path.abspath(lamp.__file__)), path
        #     os.path.dirname(os.path.abspath(__file__)), path
        )

    ext = os.path.splitext(fn)[1][1:]
    if ext in ['xls', 'xlsx']:
        df = pd.read_excel(fn, sheet_name=sheet_name, header=0)
    elif ext in ['txt', 'csv', 'tsv', 'dat']:
        df = pd.read_table(fn, header=0, sep=sep)
    else:
        raise ValueError("Data must be: xls, xlsx, txt, csv, tsv or dat.")

    # tidy up
    df = df.remove_empty().clean_names()

    # rename if possible
    if "molecular_formula" not in df.columns:
        df = df.rename(columns={"formula": "molecular_formula"})
    if "compound_name" not in df.columns:
        df = df.rename(columns={"name": "compound_name"})
    if "exact_mass" not in df.columns and "m_z" in df.columns:
        df = df.rename(columns={"m_z": "exact_mass"})
    else:
        df = df.rename(columns={"ion_m_z": "exact_mass"})

    # reference file must have either "molecular_formula" or "exact_mass",
    # or both
    if ("molecular_formula" not in df.columns) and (
            "exact_mass" not in df.columns):
        raise ValueError("Input data must have 'molecular_formula' or"
                         " 'exact_mass' columns")

    # df = df.reorder_columns(['molecular_formula', 'molecular_name'])

    if ('exact_mass' in df.columns) and (df.exact_mass.dtype != "float64"):
        df = df.assign(exact_mass=lambda x: x["exact_mass"].astype(float))

    if "ion_mode" in df and ion_mode:
        # 30-10-2024, Wed: use regex for case-insensitive and partial match
        idx = df['ion_mode'].str.contains(ion_mode, case=False)
        if idx.sum() > 0:
            df = df[idx]
        else:
            raise ValueError("Ion mode must be pos/positive or neg/negative.")

    # calculate exact mass
    if calc or ('exact_mass' not in df.columns):
        df = cal_mass(df, lib_adducts)

    return df


# -------------------------------------------------------------------------
# wl-24-12-2023, Sun: get library in data frame format
# wl-09-10-2024, Wed: set 'filename' default value
# wl-30-10-2024, Wed: minor changes
def read_lib(fn="", ion_mode="pos", sheet_name=0, sep="\t"):
    """
    Load adducts library file.

    Parameters
    ----------
    fn : str
        Full path for an library file. if empty, use default library file.
    ion_mode : str
        A string for ion mode, "pos" or "neg".
    sheet_name: str or int
        This argument is used for Excel file format. Strings are used for
        sheet names. Integers are used in zero-indexed sheet positions
        (chart sheets do not count as a sheet position).

        Available cases:

        * Defaults to ``0``: 1st sheet as a `DataFrame`
        * ``1``: 2nd sheet as a `DataFrame`
        * ``"Sheet1"``: Load sheet with name "Sheet1"

    sep : str
        File delimiter, `\t` or `,`.

    Returns
    -------
    DataFrame
        A library data frame.
    """

    # load default library
    if not fn:
        path = 'lib/adducts.txt'
        fn = os.path.join(
            os.path.dirname(os.path.abspath(lamp.__file__)), path
        #     os.path.dirname(os.path.abspath(__file__)), path
        )
        sep = "\t"

    ext = os.path.splitext(fn)[1][1:]
    if ext in ['xls', 'xlsx']:
        df = pd.read_excel(fn, sheet_name=sheet_name, header=0)
    elif ext in ['txt', 'csv', 'tsv', 'dat']:
        df = pd.read_table(fn, header=0, sep=sep)
    else:
        raise ValueError("Data must be: xls, xlsx, txt, csv, tsv or dat.")

    if "ion_mode" in df.columns and ion_mode:
        # 30-10-2024, Wed: use regex for case-insensitive and partial match
        idx = df['ion_mode'].str.contains(ion_mode, case=False)
        # idx = (df["ion_mode"] == ion_mode)
        if idx.sum() > 0:
            df = df[idx]

    df = df.drop("ion_mode", axis=1)

    return df


# --------------------------------------------------------------------------
# wl-21-09-2022, Wed: Read file consisting of peak list and data matrix.
# wl-02-09-2024, Mon: add 'dat' format for Galaxy data extension
# wl-10-10-2024, Thu: remove calculation of intensity average
def read_peak(fn, cols=[1, 2, 3, 4], sheet_name=0, sep="\t"):
    """
    Read metabolite data.

    The file must have peak-list and intensity data and must be one of file
    formats: xls, xlsx, txt, csv and tsv.

    Parameters
    ----------
    fn : str
        File name.
    cols : list
        A list indicating column index of 'name', 'mz', 'rt' and start of
        intensity data matrix.
    sheet_name: str or int
        This argument is used for Excel file format. Strings are used for
        sheet names. Integers are used in zero-indexed sheet positions
        (chart sheets do not count as a sheet position).

        Available cases:

        * Defaults to ``0``: 1st sheet as a `DataFrame`
        * ``1``: 2nd sheet as a `DataFrame`
        * ``"Sheet1"``: Load sheet with name "Sheet1"

    sep : str
        File separator for txt, csv and tsv file. `\t` or `,`.

    Returns
    -------
    DataFrame
        A data frame with the first 3 columns as name, mz and rt.
    """

    ext = os.path.splitext(fn)[1][1:]
    if ext in ['xls', 'xlsx']:
        data = pd.read_excel(fn, sheet_name=sheet_name, header=0)
    elif ext in ['txt', 'csv', 'tsv', 'dat']:
        data = pd.read_table(fn, header=0, sep=sep)
    else:
        raise ValueError("Data must be: xls, xlsx, txt, csv, tsv or dat.")

    # data = pd.read_table(fn, header=0, sep=sep)

    # get python index
    cols = [x - 1 for x in cols]
    # cols = list(map(lambda x: x - 1, cols))

    # get peak list
    peak = data.iloc[:, cols[:3]]

    # get intensity
    mat = data.iloc[:, cols[-1]:]

    # column bind ('cbind' in R)
    data = pd.concat([peak, mat], axis=1, ignore_index=False)

    data.rename(columns={data.columns[0]: 'name',
                         data.columns[1]: 'mz',
                         data.columns[2]: 'rt'},
                inplace=True)
    data['name'] = data['name'].astype(str)

    # replace 0 with na for the future correlation analysis
    data.replace(0, np.nan, inplace=True)

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# wl-07-10-2024, Mon: commence

from lamp import __version__
import argparse
import sys
import sqlite3
from PySide6.QtWidgets import QApplication
from lamp import gui
from lamp import anno
from lamp import stats


# --------------------------------------------------------------------------
def main():
    separators = {"tab": "\t", "comma": ","}

    print("Executing lamp version {}.".format(__version__))

    parser = argparse.ArgumentParser(
        description='Compounds Annotation of LC-MS data',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    subparsers = parser.add_subparsers(dest='step')
    parser_am = subparsers.add_parser('cli',
                                      help='Annotate metabolites in CLI.')
    parser_gui = subparsers.add_parser('gui',
                                       help='Annotate metabolites in GUI.')

    # ---------------------------------------------------------------------
    # data loading
    parser_am.add_argument('--input-data', type=str, required=True,
                           help="Data set including peak-list.")
    parser_am.add_argument('--col-idx', default="1,2,3,4", type=str,
                           help="Column index of name, mz, rt and start of"
                                " data intensity")
    parser_am.add_argument('--input-sep', default="tab", type=str,
                           choices=["tab", "comma"],
                           help="Values in input or output file are "
                                "separated by this character.")

    # ---------------------------------------------------------------------
    # feature grouping with correlation analysis
    parser_am.add_argument('--thres-rt', default=1.0, type=float,
                           help="Threshold of retention time difference.")
    parser_am.add_argument('--thres-corr', default=0.5, type=float,
                           help="Threshold of correlation coefficient.")
    parser_am.add_argument('--thres-pval', default=0.05, type=float,
                           help="Threshold of correlation p-values.")
    parser_am.add_argument('--method', default="pearson", type=str,
                           choices=["pearson", "spearman"],
                           help="Correlation method.")
    parser_am.add_argument('--positive', action='store_true',
                           help="Positive correlation coefficients or not.")

    # ---------------------------------------------------------------------
    # compounds annotation with reference and adduct files
    parser_am.add_argument('--ppm', default=5.0, type=float,
                           help="Mass tolerance in parts per million.")
    parser_am.add_argument('--ion-mode', default='pos', type=str,
                           choices=["pos", "neg"],
                           help="Ion mode of data set.")
    parser_am.add_argument('--ref-path', type=str, default=None,
                           required=False,
                           help="Reference file for compound matching.")
    parser_am.add_argument('--ref-sep', default="tab", type=str,
                           choices=["tab", "comma"],
                           help="Values in input or output file are "
                                "separated by this character.")
    parser_am.add_argument('--cal-mass', action="store_true",
                           help="Calculate mass based on NIST database.")
    parser_am.add_argument('--add-path', type=str, default=None,
                           required=False,
                           help="Adducts library for compound match mass"
                                " adjustment.")
    parser_am.add_argument('--add-sep', default="tab", type=str,
                           choices=["tab", "comma"],
                           help="Values in input or output file are "
                                "separated by this character.")

    # ---------------------------------------------------------------------
    # results outcome
    parser_am.add_argument('--save-db', action="store_true",
                           help="Save all results in a sql database.")
    parser_am.add_argument('--save-mr', action="store_true",
                           help="Save multiple row result.")

    parser_am.add_argument('--db-out', type=str, required=True,
                           help="All results saved in a sqlite database.")
    parser_am.add_argument('--sr-out', type=str, required=True,
                           help="Compound annotation results")
    parser_am.add_argument('--sr-sep', default="tab", type=str,
                           choices=["tab", "comma"],
                           help="Values in input or output file are "
                                "separated by this character.")
    parser_am.add_argument('--mr-out', type=str, required=True,
                           help="Compound annotation results in multiple"
                                " row format")
    parser_am.add_argument('--mr-sep', default="tab", type=str,
                           choices=["tab", "comma"],
                           help="Values in input or output file are "
                                "separated by this character.")

    # ---------------------------------------------------------------------
    args = parser.parse_args()
    print(args)

    if args.step == "cli":

        # -----------------------------------------------------------------
        # get data with peak list and intensity matrix
        idx_list = [int(item.strip()) for item in args.col_idx.split(',')]
        df = anno.read_peak(fn=args.input_data, cols=idx_list,
                            sep=separators[args.input_sep])

        # -----------------------------------------------------------------
        # load adducts library for mass adjust if mass calculation is needed
        lib_add = anno.read_lib(fn=args.add_path,
                                ion_mode=args.ion_mode,
                                sep=separators[args.add_sep])

        # -----------------------------------------------------------------
        # load reference library and calculate exact mass if needed.
        ref = anno.read_ref(fn=args.ref_path,
                            ion_mode=args.ion_mode,
                            sep=separators[args.ref_sep],
                            calc=args.cal_mass,
                            lib_adducts=lib_add)

        # -----------------------------------------------------------------
        # match compound based on exact mass
        match = anno.comp_match_mass(df, args.ppm, ref)

        # -----------------------------------------------------------------
        # correlation analysis with corr, pval and rt_diff
        corr = stats.comp_corr_rt(df,
                                  thres_rt=args.thres_rt,
                                  thres_corr=args.thres_corr,
                                  thres_pval=args.thres_pval,
                                  method=args.method,
                                  positive=args.positive)

        # get correlation group and size
        corr_df = stats.corr_grp_size(corr)

        # -----------------------------------------------------------------
        # get summary of metabolite annotation
        sr, mr = anno.comp_summ(df, match)
        # merge summery table with correlation analysis
        res = anno.comp_summ_corr(sr, corr_df)

        # -----------------------------------------------------------------
        # save all results to a sqlite database or not
        if args.save_db:
            conn = sqlite3.connect(args.db_out)
            df[["name", "mz", "rt"]].to_sql("peaklist", conn,
                                            if_exists="replace", index=False)
            corr_df.to_sql("corr_grp", conn, if_exists="replace", index=False)
            corr.to_sql("corr_pval_rt", conn, if_exists="replace", index=False)
            match.to_sql("match", conn, if_exists="replace", index=False)
            mr.to_sql("anno_mr", conn, if_exists="replace", index=False)
            res.to_sql("anno_sr", conn, if_exists="replace", index=False)

            conn.commit()
            conn.close()

        # save multiple row results or not
        if args.save_mr:
            mr.to_csv(args.mr_out, sep=separators[args.mr_sep], index=False)

        # save results
        res.to_csv(args.sr_out, sep=separators[args.sr_sep], index=False)

    if args.step == "gui":
        # Exception Handling
        try:
            app = QApplication(sys.argv)
            form = gui.lamp_app()
            form.show()
            sys.exit(app.exec())
        except NameError:
            print("Name Error:", sys.exc_info()[1])
        except SystemExit:
            print("Closing Window...")
        except Exception:
            print(sys.exc_info()[1])


# -------------------------------------------------------------------------
if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# wl-19-09-2024, Thu: commence

import os
import sys
import sqlite3
import pandas as pd
from functools import partial
from PySide6 import QtCore, QtWidgets
from lamp.qt import lamp_form
from lamp import anno
from lamp import stats
from lamp import utils


# -------------------------------------------------------------------------
# wl-19-09-2024, Thu: commence
class lamp_app(QtWidgets.QMainWindow, lamp_form.Ui_MainWindow):
    def __init__(self, *args, **kwargs):
        super(lamp_app, self).__init__(*args, **kwargs)
        self.setupUi(self)

        # ---- Input data ----
        self.pushButton_wd.clicked.connect(
            partial(self.open_directory, self.lineEdit_wd)
        )
        self.pushButton_data.clicked.connect(
            partial(self.open_file, self.lineEdit_data)
        )

        # ---- Group features ----

        # ---- Annotate compounds ----
        self.pushButton_ref.clicked.connect(
            partial(self.open_file, self.lineEdit_ref)
        )
        self.pushButton_add.clicked.connect(
            partial(self.open_file, self.lineEdit_add)
        )
        self.checkBox_mass_cal.clicked.connect(self.annotate_compounds)

        # ---- Save results ----
        self.pushButton_summ.clicked.connect(
            partial(self.save_file, self.lineEdit_summ, "anno_summ")
        )
        self.pushButton_sql.clicked.connect(
            partial(self.save_file,
                    self.lineEdit_sql, "anno_comp.db")
        )

        # ---- Others ----
        # self.path_wd = os.path.expanduser("~")
        self.path_wd = os.getcwd()

        self.pushButton_cancel.clicked.connect(
            QtCore.QCoreApplication.instance().quit
        )
        self.pushButton_start.clicked.connect(self.run)

    # ---------------------------------------------------------------------
    def annotate_compounds(self):
        if self.checkBox_mass_cal.isChecked():
            self.label_add.setEnabled(True)
            self.pushButton_add.setEnabled(True)
            self.lineEdit_add.setEnabled(True)
            self.label_lib_sep.setEnabled(True)
            self.comboBox_lib_sep.setEnabled(True)
        else:
            self.label_add.setEnabled(False)
            self.pushButton_add.setEnabled(False)
            self.lineEdit_add.setEnabled(False)
            self.label_lib_sep.setEnabled(False)
            self.comboBox_lib_sep.setEnabled(False)

    # ---------------------------------------------------------------------
    def open_directory(self, field):
        d = QtWidgets.QFileDialog.getExistingDirectory(
            None, "Select a folder", self.path_wd
        )
        if d:
            if str(d) == "":
                QtWidgets.QMessageBox.critical(
                    None,
                    "Select a folder",
                    "No folder selected",
                    QtWidgets.QMessageBox.Ok,
                )
            else:
                field.setText(d)
                self.path_wd = d
        return

    # ---------------------------------------------------------------------
    def open_file(self, field, field_extra=None):
        d = QtWidgets.QFileDialog.getOpenFileName(self, "Select File",
                                                  self.path_wd)
        if d:
            if str(d[0]) == "":
                QtWidgets.QMessageBox.critical(
                    None, "Select File", "No file selected",
                    QtWidgets.QMessageBox.Ok
                )
            else:
                field.setText(d[0])
                if field_extra and field_extra.text() == "Use default":
                    field_extra.setText(d[0])
        return

    # ---------------------------------------------------------------------
    def save_file(self, field, filename):
        d = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save File", os.path.join(self.path_wd, filename)
        )
        if d:
            if str(d[0]) == "":
                QtWidgets.QMessageBox.critical(
                    None,
                    "Save File",
                    "Provide a valid filename",
                    QtWidgets.QMessageBox.Ok,
                )
            else:
                field.setText(d[0])
        return

    # ---------------------------------------------------------------------
    def run(self):
        if not os.path.isfile(self.lineEdit_data.text()):
            QtWidgets.QMessageBox.critical(
                None,
                "Select file",
                "Select file(s) for input data (peak-list + intensity matrix)",
                QtWidgets.QMessageBox.Ok,
            )
            return

        # self.hide()
        self.pushButton_start.setEnabled(False)

        sepa = {"tab": "\t", "comma": ","}
        mode = {"Positive": "pos", "Negative": "neg"}

        # --------------------------------------------------------------
        if self.lineEdit_add.text() == "Use default":
            add_path = ""
        else:
            add_path = self.lineEdit_add.text()

        if self.lineEdit_ref.text() == "Use default":
            ref_path = ""
        else:
            ref_path = self.lineEdit_ref.text()

        # -----------------------------------------------------------------
        # get data with peak list and intensity matrix
        cols = [
            self.spinBox_name_col.value(),
            self.spinBox_mz_col.value(),
            self.spinBox_rt_col.value(),
            self.spinBox_data_col.value(),
        ]

        # -----------------------------------------------------------------
        df = anno.read_peak(self.lineEdit_data.text(), cols=cols,
                            sep=sepa[self.comboBox_data_sep.currentText()])
        lib_add = anno.read_lib(
            fn=add_path,
            ion_mode=mode[self.comboBox_ion_mode.currentText()],
            sep=sepa[self.comboBox_lib_sep.currentText()]
        )
        ref = anno.read_ref(
            fn=ref_path,
            ion_mode=mode[self.comboBox_ion_mode.currentText()],
            sep=sepa[self.comboBox_ref_sep.currentText()],
            calc=self.checkBox_mass_cal.isChecked(),
            lib_adducts=lib_add
        )
        match = anno.comp_match_mass(df, self.doubleSpinBox_ppm.value(), ref)
        print("\n***Metabolites match done***")

        # -----------------------------------------------------------------
        # correlation analysis with corr, pval and rt_diff
        if self.comboBox_method.currentText() == "Pearson correlation":
            method = "pearson"
        else:
            method = "spearman"

        utils._tic()
        corr = stats.comp_corr_rt(
            df,
            thres_rt=self.doubleSpinBox_thres_rt.value(),
            thres_corr=self.doubleSpinBox_thres_corr.value(),
            thres_pval=self.doubleSpinBox_thres_pval.value(),
            method=method,
            positive=self.checkBox_pos.isChecked()
        )
        utils._toc()

        # get correlation group and size
        utils._tic()
        corr_df = stats.corr_grp_size(corr)
        utils._toc()

        print("\n***Correlation analysis done***")

        # -----------------------------------------------------------------
        # get summary of metabolite annotation
        sr, mr = anno.comp_summ(df, match)
        # merge summery table with correlation analysis
        res = anno.comp_summ_corr(sr, corr_df)
        print("\n***Summary done***")

        # -----------------------------------------------------------------
        # save all results to a sqlite database or not
        conn = sqlite3.connect(self.lineEdit_sql.text())
        df[["name", "mz", "rt"]].to_sql("peaklist", conn,
                                        if_exists="replace", index=False)
        corr_df.to_sql("corr_grp", conn, if_exists="replace", index=False)
        corr.to_sql("corr_pval_rt", conn, if_exists="replace", index=False)
        match.to_sql("match", conn, if_exists="replace", index=False)
        mr.to_sql("anno_mr", conn, if_exists="replace", index=False)
        res.to_sql("anno_sr", conn, if_exists="replace", index=False)

        conn.commit()
        conn.close()

        if self.comboBox_ext.currentText() == "xlsx":
            xlsx_out = self.lineEdit_summ.text() + ".xlsx"
            with pd.ExcelWriter(xlsx_out, mode="w", engine="openpyxl") as writer:
                res.to_excel(writer, sheet_name="single-row", index=False)
                mr.to_excel(writer, sheet_name="multiple-row", index=False)
        elif self.comboBox_ext.currentText() == "tsv":
            tsv_out_s = self.lineEdit_summ.text() + "_s.tsv"
            tsv_out_m = self.lineEdit_summ.text() + "_m.tsv"
            res.to_csv(tsv_out_s, sep="\t", index=False)
            mr.to_csv(tsv_out_m, sep="\t", index=False)
        else:
            csv_out_s = self.lineEdit_summ.text() + "_s.csv"
            csv_out_m = self.lineEdit_summ.text() + "_m.csv"
            res.to_csv(csv_out_s, sep=",", index=False)
            mr.to_csv(csv_out_m, sep=",", index=False)

        print("\n***Save results done. You may close this app.***\n")

        self.pushButton_start.setEnabled(True)
        # self.close()


# -------------------------------------------------------------------------
# wl-31-10-2024, Thu: test and debug GUI
def main():
    # Exception Handling
    try:
        app = QtWidgets.QApplication(sys.argv)
        form = lamp_app()
        form.show()
        sys.exit(app.exec())
    except NameError:
        print("Name Error:", sys.exc_info()[1])
    except SystemExit:
        print("Closing Window...")
    except Exception:
        print(sys.exc_info()[1])


# -------------------------------------------------------------------------
# wl-31-10-2024, Thu: command line: 'python gui.py' to test
if __name__ == '__main__':
    main()



import time


# ------------------------------------------------------------------------
# wl-29-12-2023, Fri: convert 2-columns data frame to dict
def df2dict(df):
    """
    Convert 2-columns DataFrame to dict.

    Parameters
    ----------
    df : DataFrame
        a two columns data frame.

    Returns
    -------
    dict
        a dict.

    Notes
    -----
    This function is used to speedy query.
    """

    res = zip(df.iloc[:, 0], df.iloc[:, 1])
    res = list(res)
    res = dict(res)
    return res


# -------------------------------------------------------------------------
# wl-18-05-2024, Sat: flat column names.
# use 'pipe' inside a chain
def flatten_cols(df):
    """
    Flat column names.

    Parameters
    ----------
    df : DataFrame
        a data frame.

    Returns
    -------
    DataFrame
        a data frame with flattened column names.
    """

    df.columns = ["_".join(x) for x in df.columns.to_flat_index()]
    return df


# -------------------------------------------------------------------------
# wl-27-12-2023, Wed: Remove empty cell and flatten list
def flatten_list(list):
    """
    Remove empty cell and flatten list.

    Parameters
    ----------
    list : list
        a nested list with empty elements.

    Returns
    -------
    list
        a flattened list.
    """

    # remove empty item
    list = [x for x in list if x]
    # flatten list
    res = [item for lt in list for item in lt]
    return res


# -------------------------------------------------------------------------
# wl-02-08-2022, Tue: Measure running time.
# Matlab's tic and toc time measuring. From https://bit.ly/3SuXktQ
def _tic():
    """ Matlab's tic and toc time measuring. """
    global startTime_for_tictoc
    startTime_for_tictoc = time.time()


# -------------------------------------------------------------------------
def _toc():
    """ Matlab's tic and toc time measuring. """
    if 'startTime_for_tictoc' in globals():
        elap_time = time.time() - startTime_for_tictoc
        print("Elapsed time: " + str(elap_time) + " seconds.")
    else:
        print("Toc: start time not set")



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'lamp_form.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QDoubleSpinBox,
    QGridLayout, QGroupBox, QLabel, QLineEdit,
    QMainWindow, QPushButton, QScrollArea, QSizePolicy,
    QSpinBox, QVBoxLayout, QWidget)

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.setEnabled(True)
        MainWindow.resize(844, 695)
        MainWindow.setAnimated(False)
        self.actionExampleData = QAction(MainWindow)
        self.actionExampleData.setObjectName(u"actionExampleData")
        self.actionAbout = QAction(MainWindow)
        self.actionAbout.setObjectName(u"actionAbout")
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.scrollArea = QScrollArea(self.centralwidget)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setGeometry(QRect(0, 0, 841, 651))
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 839, 649))
        self.verticalLayout_2 = QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.groupBox_input = QGroupBox(self.scrollAreaWidgetContents)
        self.groupBox_input.setObjectName(u"groupBox_input")
        self.gridLayout_3 = QGridLayout(self.groupBox_input)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.gridLayout_3.setVerticalSpacing(5)
        self.gridLayout_3.setContentsMargins(10, 5, 10, 5)
        self.lineEdit_wd = QLineEdit(self.groupBox_input)
        self.lineEdit_wd.setObjectName(u"lineEdit_wd")
        self.lineEdit_wd.setEnabled(True)
        self.lineEdit_wd.setMinimumSize(QSize(150, 0))
        self.lineEdit_wd.setMaximumSize(QSize(150, 16777215))
        self.lineEdit_wd.setReadOnly(True)

        self.gridLayout_3.addWidget(self.lineEdit_wd, 4, 1, 1, 1)

        self.spinBox_mz_col = QSpinBox(self.groupBox_input)
        self.spinBox_mz_col.setObjectName(u"spinBox_mz_col")
        self.spinBox_mz_col.setMinimum(1)
        self.spinBox_mz_col.setMaximum(999999)
        self.spinBox_mz_col.setValue(2)

        self.gridLayout_3.addWidget(self.spinBox_mz_col, 7, 5, 1, 1)

        self.label_mz_col = QLabel(self.groupBox_input)
        self.label_mz_col.setObjectName(u"label_mz_col")

        self.gridLayout_3.addWidget(self.label_mz_col, 7, 4, 1, 1)

        self.spinBox_data_col = QSpinBox(self.groupBox_input)
        self.spinBox_data_col.setObjectName(u"spinBox_data_col")
        self.spinBox_data_col.setMinimum(4)
        self.spinBox_data_col.setMaximum(999999)
        self.spinBox_data_col.setValue(4)

        self.gridLayout_3.addWidget(self.spinBox_data_col, 8, 5, 1, 1)

        self.label_wd = QLabel(self.groupBox_input)
        self.label_wd.setObjectName(u"label_wd")
        self.label_wd.setEnabled(True)

        self.gridLayout_3.addWidget(self.label_wd, 4, 0, 1, 1)

        self.spinBox_name_col = QSpinBox(self.groupBox_input)
        self.spinBox_name_col.setObjectName(u"spinBox_name_col")
        self.spinBox_name_col.setMinimum(1)
        self.spinBox_name_col.setMaximum(999999)
        self.spinBox_name_col.setValue(1)

        self.gridLayout_3.addWidget(self.spinBox_name_col, 7, 1, 1, 1)

        self.label_rt_col = QLabel(self.groupBox_input)
        self.label_rt_col.setObjectName(u"label_rt_col")

        self.gridLayout_3.addWidget(self.label_rt_col, 8, 0, 1, 1)

        self.label_name_col = QLabel(self.groupBox_input)
        self.label_name_col.setObjectName(u"label_name_col")

        self.gridLayout_3.addWidget(self.label_name_col, 7, 0, 1, 1)

        self.pushButton_wd = QPushButton(self.groupBox_input)
        self.pushButton_wd.setObjectName(u"pushButton_wd")
        self.pushButton_wd.setEnabled(True)
        self.pushButton_wd.setMinimumSize(QSize(95, 23))
        self.pushButton_wd.setMaximumSize(QSize(95, 16777215))

        self.gridLayout_3.addWidget(self.pushButton_wd, 4, 2, 1, 1)

        self.label_data_col = QLabel(self.groupBox_input)
        self.label_data_col.setObjectName(u"label_data_col")

        self.gridLayout_3.addWidget(self.label_data_col, 8, 4, 1, 1)

        self.label_input = QLabel(self.groupBox_input)
        self.label_input.setObjectName(u"label_input")
        self.label_input.setEnabled(True)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_input.sizePolicy().hasHeightForWidth())
        self.label_input.setSizePolicy(sizePolicy)
        font = QFont()
        font.setPointSize(13)
        font.setBold(True)
        self.label_input.setFont(font)

        self.gridLayout_3.addWidget(self.label_input, 0, 0, 1, 1)

        self.spinBox_rt_col = QSpinBox(self.groupBox_input)
        self.spinBox_rt_col.setObjectName(u"spinBox_rt_col")
        self.spinBox_rt_col.setMinimum(1)
        self.spinBox_rt_col.setMaximum(999999)
        self.spinBox_rt_col.setValue(3)

        self.gridLayout_3.addWidget(self.spinBox_rt_col, 8, 1, 1, 1)

        self.label_data = QLabel(self.groupBox_input)
        self.label_data.setObjectName(u"label_data")
        self.label_data.setEnabled(True)

        self.gridLayout_3.addWidget(self.label_data, 6, 0, 1, 1)

        self.lineEdit_data = QLineEdit(self.groupBox_input)
        self.lineEdit_data.setObjectName(u"lineEdit_data")
        self.lineEdit_data.setEnabled(True)
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.lineEdit_data.sizePolicy().hasHeightForWidth())
        self.lineEdit_data.setSizePolicy(sizePolicy1)
        self.lineEdit_data.setMinimumSize(QSize(150, 0))
        self.lineEdit_data.setMaximumSize(QSize(150, 16777215))
        self.lineEdit_data.setReadOnly(True)

        self.gridLayout_3.addWidget(self.lineEdit_data, 6, 1, 1, 1)

        self.pushButton_data = QPushButton(self.groupBox_input)
        self.pushButton_data.setObjectName(u"pushButton_data")
        self.pushButton_data.setEnabled(True)
        self.pushButton_data.setMinimumSize(QSize(95, 23))
        self.pushButton_data.setMaximumSize(QSize(95, 16777215))

        self.gridLayout_3.addWidget(self.pushButton_data, 6, 2, 1, 1)

        self.label_data_sep = QLabel(self.groupBox_input)
        self.label_data_sep.setObjectName(u"label_data_sep")

        self.gridLayout_3.addWidget(self.label_data_sep, 6, 4, 1, 1)

        self.comboBox_data_sep = QComboBox(self.groupBox_input)
        self.comboBox_data_sep.addItem("")
        self.comboBox_data_sep.addItem("")
        self.comboBox_data_sep.setObjectName(u"comboBox_data_sep")

        self.gridLayout_3.addWidget(self.comboBox_data_sep, 6, 5, 1, 1)


        self.verticalLayout_2.addWidget(self.groupBox_input)

        self.groupBox_group = QGroupBox(self.scrollAreaWidgetContents)
        self.groupBox_group.setObjectName(u"groupBox_group")
        self.gridLayout_4 = QGridLayout(self.groupBox_group)
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.gridLayout_4.setHorizontalSpacing(6)
        self.gridLayout_4.setVerticalSpacing(5)
        self.gridLayout_4.setContentsMargins(10, 5, 10, 5)
        self.checkBox_pos = QCheckBox(self.groupBox_group)
        self.checkBox_pos.setObjectName(u"checkBox_pos")
        self.checkBox_pos.setEnabled(True)
        self.checkBox_pos.setChecked(True)

        self.gridLayout_4.addWidget(self.checkBox_pos, 12, 0, 1, 1)

        self.doubleSpinBox_thres_corr = QDoubleSpinBox(self.groupBox_group)
        self.doubleSpinBox_thres_corr.setObjectName(u"doubleSpinBox_thres_corr")
        self.doubleSpinBox_thres_corr.setMaximum(1.000000000000000)
        self.doubleSpinBox_thres_corr.setSingleStep(0.100000000000000)
        self.doubleSpinBox_thres_corr.setValue(0.700000000000000)

        self.gridLayout_4.addWidget(self.doubleSpinBox_thres_corr, 10, 1, 1, 1)

        self.label_thres_corr = QLabel(self.groupBox_group)
        self.label_thres_corr.setObjectName(u"label_thres_corr")
        self.label_thres_corr.setEnabled(True)

        self.gridLayout_4.addWidget(self.label_thres_corr, 10, 0, 1, 1)

        self.label_thres_pval = QLabel(self.groupBox_group)
        self.label_thres_pval.setObjectName(u"label_thres_pval")
        self.label_thres_pval.setEnabled(True)

        self.gridLayout_4.addWidget(self.label_thres_pval, 10, 2, 1, 1)

        self.label_group = QLabel(self.groupBox_group)
        self.label_group.setObjectName(u"label_group")
        self.label_group.setFont(font)

        self.gridLayout_4.addWidget(self.label_group, 1, 0, 1, 1)

        self.doubleSpinBox_thres_pval = QDoubleSpinBox(self.groupBox_group)
        self.doubleSpinBox_thres_pval.setObjectName(u"doubleSpinBox_thres_pval")
        self.doubleSpinBox_thres_pval.setDecimals(10)
        self.doubleSpinBox_thres_pval.setMaximum(1.000000000000000)
        self.doubleSpinBox_thres_pval.setSingleStep(0.010000000000000)
        self.doubleSpinBox_thres_pval.setValue(0.010000000000000)

        self.gridLayout_4.addWidget(self.doubleSpinBox_thres_pval, 10, 3, 1, 1)

        self.label_thres_rt = QLabel(self.groupBox_group)
        self.label_thres_rt.setObjectName(u"label_thres_rt")
        self.label_thres_rt.setEnabled(True)

        self.gridLayout_4.addWidget(self.label_thres_rt, 3, 2, 1, 1)

        self.doubleSpinBox_thres_rt = QDoubleSpinBox(self.groupBox_group)
        self.doubleSpinBox_thres_rt.setObjectName(u"doubleSpinBox_thres_rt")
        self.doubleSpinBox_thres_rt.setMaximum(9999999.000000000000000)
        self.doubleSpinBox_thres_rt.setValue(1.000000000000000)

        self.gridLayout_4.addWidget(self.doubleSpinBox_thres_rt, 3, 3, 1, 1)

        self.comboBox_method = QComboBox(self.groupBox_group)
        self.comboBox_method.addItem("")
        self.comboBox_method.addItem("")
        self.comboBox_method.setObjectName(u"comboBox_method")

        self.gridLayout_4.addWidget(self.comboBox_method, 3, 1, 1, 1)

        self.label_method = QLabel(self.groupBox_group)
        self.label_method.setObjectName(u"label_method")
        self.label_method.setEnabled(True)
        self.label_method.setIndent(-1)

        self.gridLayout_4.addWidget(self.label_method, 3, 0, 1, 1)


        self.verticalLayout_2.addWidget(self.groupBox_group)

        self.groupBox_annotate = QGroupBox(self.scrollAreaWidgetContents)
        self.groupBox_annotate.setObjectName(u"groupBox_annotate")
        self.groupBox_annotate.setEnabled(True)
        self.gridLayout = QGridLayout(self.groupBox_annotate)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setVerticalSpacing(5)
        self.gridLayout.setContentsMargins(10, 5, 10, 5)
        self.checkBox_mass_cal = QCheckBox(self.groupBox_annotate)
        self.checkBox_mass_cal.setObjectName(u"checkBox_mass_cal")
        self.checkBox_mass_cal.setEnabled(True)
        self.checkBox_mass_cal.setChecked(False)

        self.gridLayout.addWidget(self.checkBox_mass_cal, 11, 0, 1, 1)

        self.doubleSpinBox_ppm = QDoubleSpinBox(self.groupBox_annotate)
        self.doubleSpinBox_ppm.setObjectName(u"doubleSpinBox_ppm")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.doubleSpinBox_ppm.sizePolicy().hasHeightForWidth())
        self.doubleSpinBox_ppm.setSizePolicy(sizePolicy2)
        self.doubleSpinBox_ppm.setMaximum(100000.000000000000000)
        self.doubleSpinBox_ppm.setValue(5.000000000000000)

        self.gridLayout.addWidget(self.doubleSpinBox_ppm, 6, 3, 1, 1)

        self.comboBox_ref_sep = QComboBox(self.groupBox_annotate)
        self.comboBox_ref_sep.addItem("")
        self.comboBox_ref_sep.addItem("")
        self.comboBox_ref_sep.setObjectName(u"comboBox_ref_sep")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.comboBox_ref_sep.sizePolicy().hasHeightForWidth())
        self.comboBox_ref_sep.setSizePolicy(sizePolicy3)

        self.gridLayout.addWidget(self.comboBox_ref_sep, 9, 4, 1, 1)

        self.label_annotate = QLabel(self.groupBox_annotate)
        self.label_annotate.setObjectName(u"label_annotate")
        self.label_annotate.setFont(font)

        self.gridLayout.addWidget(self.label_annotate, 0, 0, 1, 1)

        self.label_ref_sep = QLabel(self.groupBox_annotate)
        self.label_ref_sep.setObjectName(u"label_ref_sep")

        self.gridLayout.addWidget(self.label_ref_sep, 9, 3, 1, 1)

        self.comboBox_lib_sep = QComboBox(self.groupBox_annotate)
        self.comboBox_lib_sep.addItem("")
        self.comboBox_lib_sep.addItem("")
        self.comboBox_lib_sep.setObjectName(u"comboBox_lib_sep")
        self.comboBox_lib_sep.setEnabled(False)
        sizePolicy3.setHeightForWidth(self.comboBox_lib_sep.sizePolicy().hasHeightForWidth())
        self.comboBox_lib_sep.setSizePolicy(sizePolicy3)

        self.gridLayout.addWidget(self.comboBox_lib_sep, 12, 4, 1, 1)

        self.label_ppm = QLabel(self.groupBox_annotate)
        self.label_ppm.setObjectName(u"label_ppm")
        self.label_ppm.setEnabled(True)

        self.gridLayout.addWidget(self.label_ppm, 6, 2, 1, 1)

        self.label_ion_mode = QLabel(self.groupBox_annotate)
        self.label_ion_mode.setObjectName(u"label_ion_mode")
        self.label_ion_mode.setMaximumSize(QSize(75, 16777215))

        self.gridLayout.addWidget(self.label_ion_mode, 6, 0, 1, 1)

        self.lineEdit_add = QLineEdit(self.groupBox_annotate)
        self.lineEdit_add.setObjectName(u"lineEdit_add")
        self.lineEdit_add.setEnabled(False)
        self.lineEdit_add.setMinimumSize(QSize(150, 0))
        self.lineEdit_add.setMaximumSize(QSize(150, 16777215))
        self.lineEdit_add.setReadOnly(True)

        self.gridLayout.addWidget(self.lineEdit_add, 12, 1, 1, 1)

        self.label_ref = QLabel(self.groupBox_annotate)
        self.label_ref.setObjectName(u"label_ref")

        self.gridLayout.addWidget(self.label_ref, 9, 0, 1, 1)

        self.comboBox_ion_mode = QComboBox(self.groupBox_annotate)
        self.comboBox_ion_mode.addItem("")
        self.comboBox_ion_mode.addItem("")
        self.comboBox_ion_mode.setObjectName(u"comboBox_ion_mode")
        self.comboBox_ion_mode.setMinimumSize(QSize(100, 0))
        self.comboBox_ion_mode.setMaximumSize(QSize(100, 16777215))

        self.gridLayout.addWidget(self.comboBox_ion_mode, 6, 1, 1, 1)

        self.pushButton_ref = QPushButton(self.groupBox_annotate)
        self.pushButton_ref.setObjectName(u"pushButton_ref")
        self.pushButton_ref.setEnabled(True)
        self.pushButton_ref.setMinimumSize(QSize(95, 23))
        self.pushButton_ref.setMaximumSize(QSize(95, 16777215))

        self.gridLayout.addWidget(self.pushButton_ref, 9, 2, 1, 1)

        self.label_lib_sep = QLabel(self.groupBox_annotate)
        self.label_lib_sep.setObjectName(u"label_lib_sep")
        self.label_lib_sep.setEnabled(False)

        self.gridLayout.addWidget(self.label_lib_sep, 12, 3, 1, 1)

        self.label_add = QLabel(self.groupBox_annotate)
        self.label_add.setObjectName(u"label_add")
        self.label_add.setEnabled(False)

        self.gridLayout.addWidget(self.label_add, 12, 0, 1, 1)

        self.pushButton_add = QPushButton(self.groupBox_annotate)
        self.pushButton_add.setObjectName(u"pushButton_add")
        self.pushButton_add.setEnabled(False)
        self.pushButton_add.setMinimumSize(QSize(95, 23))
        self.pushButton_add.setMaximumSize(QSize(95, 16777215))

        self.gridLayout.addWidget(self.pushButton_add, 12, 2, 1, 1)

        self.lineEdit_ref = QLineEdit(self.groupBox_annotate)
        self.lineEdit_ref.setObjectName(u"lineEdit_ref")
        self.lineEdit_ref.setEnabled(True)
        sizePolicy1.setHeightForWidth(self.lineEdit_ref.sizePolicy().hasHeightForWidth())
        self.lineEdit_ref.setSizePolicy(sizePolicy1)
        self.lineEdit_ref.setMinimumSize(QSize(120, 0))
        self.lineEdit_ref.setMaximumSize(QSize(120, 16777215))
        self.lineEdit_ref.setReadOnly(True)

        self.gridLayout.addWidget(self.lineEdit_ref, 9, 1, 1, 1)


        self.verticalLayout_2.addWidget(self.groupBox_annotate)

        self.groupBox_save = QGroupBox(self.scrollAreaWidgetContents)
        self.groupBox_save.setObjectName(u"groupBox_save")
        self.gridLayout_2 = QGridLayout(self.groupBox_save)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setVerticalSpacing(5)
        self.gridLayout_2.setContentsMargins(10, 5, 10, 5)
        self.label_ext = QLabel(self.groupBox_save)
        self.label_ext.setObjectName(u"label_ext")
        self.label_ext.setEnabled(True)

        self.gridLayout_2.addWidget(self.label_ext, 3, 6, 1, 1)

        self.lineEdit_sql = QLineEdit(self.groupBox_save)
        self.lineEdit_sql.setObjectName(u"lineEdit_sql")
        self.lineEdit_sql.setEnabled(True)
        self.lineEdit_sql.setMinimumSize(QSize(150, 0))
        self.lineEdit_sql.setMaximumSize(QSize(150, 16777215))
        self.lineEdit_sql.setReadOnly(True)

        self.gridLayout_2.addWidget(self.lineEdit_sql, 9, 1, 1, 1)

        self.pushButton_summ = QPushButton(self.groupBox_save)
        self.pushButton_summ.setObjectName(u"pushButton_summ")
        self.pushButton_summ.setMinimumSize(QSize(63, 23))

        self.gridLayout_2.addWidget(self.pushButton_summ, 3, 3, 1, 1)

        self.label_save = QLabel(self.groupBox_save)
        self.label_save.setObjectName(u"label_save")
        self.label_save.setFont(font)

        self.gridLayout_2.addWidget(self.label_save, 1, 0, 1, 1)

        self.label_sql = QLabel(self.groupBox_save)
        self.label_sql.setObjectName(u"label_sql")
        self.label_sql.setEnabled(True)

        self.gridLayout_2.addWidget(self.label_sql, 9, 0, 1, 1)

        self.label_summ = QLabel(self.groupBox_save)
        self.label_summ.setObjectName(u"label_summ")
        self.label_summ.setEnabled(True)

        self.gridLayout_2.addWidget(self.label_summ, 3, 0, 1, 1)

        self.pushButton_sql = QPushButton(self.groupBox_save)
        self.pushButton_sql.setObjectName(u"pushButton_sql")
        self.pushButton_sql.setEnabled(True)
        self.pushButton_sql.setMinimumSize(QSize(95, 23))
        self.pushButton_sql.setMaximumSize(QSize(95, 16777215))

        self.gridLayout_2.addWidget(self.pushButton_sql, 9, 3, 1, 1)

        self.lineEdit_summ = QLineEdit(self.groupBox_save)
        self.lineEdit_summ.setObjectName(u"lineEdit_summ")
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Fixed)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.lineEdit_summ.sizePolicy().hasHeightForWidth())
        self.lineEdit_summ.setSizePolicy(sizePolicy4)
        self.lineEdit_summ.setReadOnly(True)

        self.gridLayout_2.addWidget(self.lineEdit_summ, 3, 1, 1, 1)

        self.comboBox_ext = QComboBox(self.groupBox_save)
        self.comboBox_ext.addItem("")
        self.comboBox_ext.addItem("")
        self.comboBox_ext.addItem("")
        self.comboBox_ext.setObjectName(u"comboBox_ext")
        self.comboBox_ext.setEditable(False)

        self.gridLayout_2.addWidget(self.comboBox_ext, 3, 7, 1, 1)


        self.verticalLayout_2.addWidget(self.groupBox_save)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.pushButton_cancel = QPushButton(self.centralwidget)
        self.pushButton_cancel.setObjectName(u"pushButton_cancel")
        self.pushButton_cancel.setGeometry(QRect(670, 660, 71, 31))
        self.pushButton_cancel.setMaximumSize(QSize(16777204, 16777215))
        self.pushButton_start = QPushButton(self.centralwidget)
        self.pushButton_start.setObjectName(u"pushButton_start")
        self.pushButton_start.setGeometry(QRect(760, 660, 61, 31))
        MainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(MainWindow)

        self.comboBox_ref_sep.setCurrentIndex(1)
        self.comboBox_ext.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"LAMP - Liverpool Annotation of metabolites using Mass sPectrometry", None))
        self.actionExampleData.setText(QCoreApplication.translate("MainWindow", u"Add example data", None))
        self.actionAbout.setText(QCoreApplication.translate("MainWindow", u"About", None))
        self.groupBox_input.setTitle("")
        self.lineEdit_wd.setText("")
        self.label_mz_col.setText(QCoreApplication.translate("MainWindow", u"Index of m/z value:", None))
        self.label_wd.setText(QCoreApplication.translate("MainWindow", u"Working directory:", None))
        self.label_rt_col.setText(QCoreApplication.translate("MainWindow", u"Index of retention time:", None))
        self.label_name_col.setText(QCoreApplication.translate("MainWindow", u"Index of name:", None))
        self.pushButton_wd.setText(QCoreApplication.translate("MainWindow", u"Browse...", None))
        self.label_data_col.setText(QCoreApplication.translate("MainWindow", u"Index of data starting:", None))
        self.label_input.setText(QCoreApplication.translate("MainWindow", u"Input Data", None))
        self.label_data.setText(QCoreApplication.translate("MainWindow", u"Input  data:", None))
        self.lineEdit_data.setText("")
        self.pushButton_data.setText(QCoreApplication.translate("MainWindow", u"Browse...", None))
        self.label_data_sep.setText(QCoreApplication.translate("MainWindow", u"Separator:", None))
        self.comboBox_data_sep.setItemText(0, QCoreApplication.translate("MainWindow", u"tab", None))
        self.comboBox_data_sep.setItemText(1, QCoreApplication.translate("MainWindow", u"comma", None))

        self.groupBox_group.setTitle("")
        self.checkBox_pos.setText(QCoreApplication.translate("MainWindow", u"Positive correlation", None))
        self.label_thres_corr.setText(QCoreApplication.translate("MainWindow", u"Coefficent threshold:", None))
        self.label_thres_pval.setText(QCoreApplication.translate("MainWindow", u"P-value threshold:", None))
        self.label_group.setText(QCoreApplication.translate("MainWindow", u"Group Features", None))
        self.label_thres_rt.setText(QCoreApplication.translate("MainWindow", u"RT difference threshold:", None))
        self.comboBox_method.setItemText(0, QCoreApplication.translate("MainWindow", u"Pearson correlation", None))
        self.comboBox_method.setItemText(1, QCoreApplication.translate("MainWindow", u"Spearman-rank correlation", None))

        self.label_method.setText(QCoreApplication.translate("MainWindow", u"Grouping method:", None))
        self.groupBox_annotate.setTitle("")
        self.checkBox_mass_cal.setText(QCoreApplication.translate("MainWindow", u"Calculate mass for reference", None))
        self.comboBox_ref_sep.setItemText(0, QCoreApplication.translate("MainWindow", u"tab", None))
        self.comboBox_ref_sep.setItemText(1, QCoreApplication.translate("MainWindow", u"comma", None))

        self.label_annotate.setText(QCoreApplication.translate("MainWindow", u"Annotate Coumpounds / Motabolites", None))
        self.label_ref_sep.setText(QCoreApplication.translate("MainWindow", u"Separator:", None))
        self.comboBox_lib_sep.setItemText(0, QCoreApplication.translate("MainWindow", u"tab", None))
        self.comboBox_lib_sep.setItemText(1, QCoreApplication.translate("MainWindow", u"comma", None))

        self.label_ppm.setText(QCoreApplication.translate("MainWindow", u"Mass tolerance (ppm):", None))
        self.label_ion_mode.setText(QCoreApplication.translate("MainWindow", u"Ion mode:", None))
        self.lineEdit_add.setText(QCoreApplication.translate("MainWindow", u"Use default", None))
        self.label_ref.setText(QCoreApplication.translate("MainWindow", u"Reference File:", None))
        self.comboBox_ion_mode.setItemText(0, QCoreApplication.translate("MainWindow", u"Positive", None))
        self.comboBox_ion_mode.setItemText(1, QCoreApplication.translate("MainWindow", u"Negative", None))

        self.pushButton_ref.setText(QCoreApplication.translate("MainWindow", u"Browse...", None))
        self.label_lib_sep.setText(QCoreApplication.translate("MainWindow", u"Separator:", None))
        self.label_add.setText(QCoreApplication.translate("MainWindow", u"Adduct library for mass adjustment:", None))
        self.pushButton_add.setText(QCoreApplication.translate("MainWindow", u"Browse...", None))
        self.lineEdit_ref.setText(QCoreApplication.translate("MainWindow", u"Use default", None))
        self.groupBox_save.setTitle("")
        self.label_ext.setText(QCoreApplication.translate("MainWindow", u"File Format:", None))
        self.lineEdit_sql.setText("")
        self.pushButton_summ.setText(QCoreApplication.translate("MainWindow", u"Save as...", None))
        self.label_save.setText(QCoreApplication.translate("MainWindow", u"Save Results", None))
        self.label_sql.setText(QCoreApplication.translate("MainWindow", u"SQLite database:", None))
        self.label_summ.setText(QCoreApplication.translate("MainWindow", u"Summary:", None))
        self.pushButton_sql.setText(QCoreApplication.translate("MainWindow", u"Save as...", None))
        self.lineEdit_summ.setText("")
        self.comboBox_ext.setItemText(0, QCoreApplication.translate("MainWindow", u"xlsx", None))
        self.comboBox_ext.setItemText(1, QCoreApplication.translate("MainWindow", u"tsv", None))
        self.comboBox_ext.setItemText(2, QCoreApplication.translate("MainWindow", u"csv", None))

        self.comboBox_ext.setCurrentText(QCoreApplication.translate("MainWindow", u"xlsx", None))
        self.pushButton_cancel.setText(QCoreApplication.translate("MainWindow", u"Cancel", None))
        self.pushButton_start.setText(QCoreApplication.translate("MainWindow", u"Start", None))
    # retranslateUi




# wl-08-10-2024, Tue: LAMP example
from lamp import anno, stats, utils

# =========================================================================
# data set
d_data = "./data/df_pos_3.tsv"
cols = [1, 2, 3, 4]

df = anno.read_peak(d_data, cols, sep='\t')
df

# ========================================================================
# metabolite annotation: match the exact mass
ppm = 5.0
ion_mode = "pos"
add_path = ""
ref_path = ""
# ref_path = "./data/hmdb_urine_v4_0_20200910_v1.tsv"
# ref_path = "./data/kegg_full_20210111_v1.tsv"
# ref_path = "./data/biocyc_chlamycyc_20180702_v1.tsv"
cal_mass = True
add_mass = False

# load reference library
ref = anno.read_ref(ref_path, ion_mode, calc=cal_mass)
ref

if add_mass:
    # match compounds with adduct library mass adjustment
    lib_add = anno.read_lib(add_path, ion_mode)
    match = anno.comp_match_mass_add(df, ppm, ref, lib_add)
else:
    # match compounds without adduct library mass adjustment
    match = anno.comp_match_mass(df, ppm, ref)
match
match.columns

# ========================================================================
# correlation analysis with corr, pval and rt_diff
thres_rt = 1.0
thres_corr = 0.7
thres_pval = 0.05
method = "spearman"   # "pearson"
positive = True

utils._tic()
corr = stats.comp_corr_rt(df, thres_rt, thres_corr, thres_pval, method,
                          positive)
utils._toc()
corr

# get correlation group and size
corr_df = stats.corr_grp_size(corr)
corr_df

# ========================================================================
# get summary of metabolite annotation
sr, mr = anno.comp_summ(df, match)
sr
mr

# merge summery table with correlation analysis
res = anno.comp_summ_corr(sr, corr_df)
res
